#include <algorithm>
#include <array>
#include <chrono>
#include <cmath>
#include <cstdlib>
#include <iomanip>
#include <iostream>
#include <fstream>
#include <random>
#include <string>
#include <stdexcept>
#include <queue>
#include <tuple>
#include <vector>
#include <stdint.h>

using namespace std;
using namespace std::chrono;

struct Point {
	float x, y;
};

#pragma omp declare simd notinbranch
float dist(const Point& p1, const Point& p2) {
	return sqrt((p1.x-p2.x)*(p1.x-p2.x)+(p1.y-p2.y)*(p1.y-p2.y));
}

#pragma omp declare simd notinbranch
float cost(const vector<Point>& points, int i, int j, int k) {
	const Point& p1 = points[i],& p2 = points[j],& p3 = points[k];
	return dist(p1, p2)+dist(p2, p3)+dist(p3, p1);
}

vector<tuple<int, int, int>> findTriangles(const vector<Point>& points, const vector<vector<float>>& table, const vector<vector<int>>& bestK)	{

	size_t n = table.size();
	vector<tuple<int, int, int>> triangles;
	queue<tuple<int,int>> toResove;
	toResove.emplace(0, n-1);

	while (!toResove.empty())	{
		int i, j;
		tie(i, j) = toResove.front();
		toResove.pop();
		int k = bestK[i][j];
		triangles.emplace_back(i, k, j);
		if (j > k+1)
			toResove.emplace(k, j);
		if (k > i+1)
			toResove.emplace(i, k);
	}

	return triangles;
}

tuple<vector<tuple<int, int, int>>, float> triangulate(vector<Point> points) {
	uint32_t n = points.size();
	if (n < 3)
		return make_tuple(vector<tuple<int,int,int>>(), 0.0f);

	vector<vector<int>> bestK(n, vector<int>(n));
	vector<vector<float>> table(n, vector<float>(n));

	for (int gap = 2; gap < n; gap++) {
		#pragma omp parallel for schedule(static)
		for (int j = gap; j < n; ++j)	{
			int i = j-gap;
			float vals[n];

			#pragma omp simd
			for (int k = i+1; k < j; k++)
				vals[k] = table[i][k]+table[j][k]+cost(points,i,j,k);

			const auto& p = min_element(vals+i+1, vals+j);
			table[i][j] = table[j][i] = *p;
			bestK[i][j] = (p-vals);
		}
	}

	// Find the triangles...
	float bestCost = table[0][n-1];
	vector<tuple<int, int, int>> triangles = findTriangles(points, table, bestK);
	return make_tuple(move(triangles), bestCost);
}

vector<Point> generatePoints(uint32_t n)	{
	constexpr float pi = 3.141592653589;
	float step = 2*pi/n;
	vector<Point> points;
	for (uint32_t i = 0; i < n; ++i)	{
		float angle = step*i;
		points.push_back({800.0f*cosf(angle)+800.0f, 800.0f*sinf(angle)+800.0f});
	}

	return points;
}

vector<Point> readProblem(const string& inputFile)	{
	vector<Point> points;
	ifstream bin(inputFile.c_str(), ifstream::binary);
	if (bin)	{
		int32_t n = 0;
		bin.read((char*) &n, sizeof(int32_t));
		for (uint64_t p = 0; p < n; ++p)	{
			float x = 0.0, y = 0.0;
			bin.read((char*) &x, sizeof(float));
			bin.read((char*) &y, sizeof(float));
			points.push_back({x, y});
		}

		bin.close();
	} else {
		throw invalid_argument("Cannot open the input file '"+inputFile+"' to read the problem.");
	}

	return points;
}

void writeResult(float triagCost, const vector<tuple<int, int, int>>& triangles, const string& resultFile)	{
	ofstream bout(resultFile.c_str(), ofstream::binary | ofstream::trunc);
	if (bout)	{
		bout.write((char*) &triagCost, sizeof(float));
		for (const auto& triangle : triangles)	{
			int32_t p1, p2, p3;
			tie(p1, p2, p3) = triangle;
			bout.write((char*) &p1, sizeof(int32_t));
			bout.write((char*) &p2, sizeof(int32_t));
			bout.write((char*) &p3, sizeof(int32_t));
		}

		bout.close();
	} else {
		throw invalid_argument("Cannot write the results, check the permissions.");
	}
}

void writeImage(const vector<Point>& points, const vector<tuple<int, int, int>>& triangles, const string& imageFilename)	{
	constexpr uint32_t numOfColors = 10;
	array<string, numOfColors> colors = {
		"orange", "brown", "purple", "blue", "darksalmon", "yellow", "green", "red", "lime", "aqua"
	};

	float minX = 10e6, maxX = -10e6, minY = 10e6, maxY = -10e6;
	for (const Point& p : points)	{
		minX = min(minX, p.x);
		maxX = max(maxX, p.x);
		minY = min(minY, p.y);
		maxY = max(maxY, p.y);
	}

	float mult = 1600.0f/(maxX-minX), height = ceil(mult*(maxY-minY));

	ofstream im(imageFilename);
	if (im)	{
		im<<"<svg xmlns=\"http://www.w3.org/2000/svg\" width=\"1600\" height=\""<<height<<"\">"<<endl;

		default_random_engine generator;
		uniform_int_distribution<uint32_t> colorIdx(0u, numOfColors-1u);
		for (const tuple<int, int, int>& t : triangles)	{
			int i, j, k;
			tie(i, j, k) = t;

			array<Point, 3> p { points[i], points[j], points[k] };
			for (uint32_t i = 0; i < 3; ++i)	{
				p[i].x = mult*(p[i].x-minX);
				p[i].y = mult*(p[i].y-minY);
			}

			im<<"\t<polygon points=\""<<p[0].x<<","<<p[0].y<<" "<<p[1].x<<","<<p[1].y<<" "<<p[2].x<<","<<p[2].y<<"\" ";
			im<<"style=\"fill:"<<colors[colorIdx(generator)]<<";stroke:black;stroke-width=0.3\"/>"<<endl;
		}

		for (uint32_t i = 0; i < points.size(); ++i)	{
			array<Point, 2> p { points[i % points.size()], points[(i+1) % points.size()] };
			for (uint32_t i = 0; i < 2; ++i)	{
				p[i].x = mult*(p[i].x-minX);
				p[i].y = mult*(p[i].y-minY);
			}	

			im<<"\t<line x1=\""<<p[0].x<<"\" y1=\""<<p[0].y<<"\" x2=\""<<p[1].x<<"\" y2=\""<<p[1].y<<"\" ";
			im<<"stroke-width=\"2\" stroke=\"black\"/>"<<endl;
		}

		im<<"</svg>"<<endl;
		im.close();
	} else {
		cerr<<"Cannot write the result to svg file!"<<endl;
	}
}

int main(int argc, char* argv[]) {
	uint32_t numberOfPoints = 32u;
	string imageFilename, inputFile, resultFile;
	for (int i = 1; i < argc; ++i)	{
		string parg = argv[i];
		if (parg == "--help" || parg == "-h")	{
			cout<<"Triangulation of a convex polygon."<<endl;
			cout<<endl<<"Usage:"<<endl;
			cout<<"\t"<<argv[0]<<" input_problem.bin result.bin [options]"<<endl<<endl;
			cout<<"General options:"<<endl;
			cout<<"\t--output-image ARG, -of ARG, ARG: {FILE_NAME}.svg"<<endl;
			cout<<"\t\tThe output svg file demonstrating the triangulation."<<endl;
			cout<<"\t--help, -h"<<endl;
			cout<<"\t\tIt prints this help."<<endl;
			return 0;
		}

		if (parg == "--output-image" || parg == "-of")	{
			if (i+1 < argc)	{
				imageFilename = argv[++i];
			} else {
				cerr<<"Expected a filename for the output image!"<<endl;
				return 1;
			}
		}

		if (!parg.empty() && parg[0] != '-')	{
			if (inputFile.empty())
				inputFile = parg;
			else
				resultFile = parg;
		}
	}

	try {
		high_resolution_clock::time_point start = high_resolution_clock::now();

		float criterion;
		vector<tuple<int, int, int>> triangles;
		const vector<Point>& points = readProblem(inputFile);

		tie(triangles, criterion) = triangulate(points);
		double totalDuration = duration_cast<duration<double>>(high_resolution_clock::now()-start).count();

		if (!resultFile.empty())
			writeResult(criterion, triangles, resultFile);

		if (!imageFilename.empty())
			writeImage(points, triangles, imageFilename);

		cout<<"Cost of triangulation: "<<criterion<<endl;
		cout<<"computational time: "<<totalDuration<<" s"<<endl;

	} catch (exception& e) {
		cerr<<"Exception caught: "<<e.what()<<endl;
		return 2;
	}

	return 0;
}

